#include "I2CDisplay.h"

void sleep (uint8_t ms) {
    for(; ms > 0; ms--) _delay_ms(1);
}

void i2cdisplay_init(uint8_t lcd_addr, uint8_t lcd_cols, uint8_t lcd_rows, uint8_t font_type) {
    addr = lcd_addr;

    cols = lcd_cols;
    rows = lcd_rows;
    display_backlight_mask = LCD_NOBACKLIGHT;

    // We never know in which state the display might be, therefore
    // a complete reset has to be performed. Note:
    // http://electronics.stackexchange.com/questions/26120/hd44780-how-long-do-i-have-to-wait-for-busy-flag-to-reset

    // according to datasheet, we need at least 40ms after power rises above 2.7V
    // before sending commands. 
    sleep(100);

    // first ensure we're in 8-bit mode 

    // maybe currently in 4-bit mode. Ensure no second nibble pending...
    i2cdisplay_write4bits(0x03);
    sleep(5);

    // maybe currently in 4-bit mode, switch to 8-bit (first nibble)
    i2cdisplay_write4bits(0x03);
    _delay_us(100);

    // maybe currently in 4-bit mode, switch to 8-bit (second nibble)
    i2cdisplay_write4bits(0x03); 
    _delay_us(100);
    
    // finally, we are in 8-bit mode, now set to 4-bit mode
    i2cdisplay_write4bits(0x02); 
    _delay_us(100);

    // now proper initialization over 4-bit interface can be done
    // set # lines and font size.
    // Values can only be set once, so no global variable is needed
    uint8_t function = LCD_4BITMODE | font_type;
    if (rows > 1)
        function |= LCD_2LINE;
    else 
        function |= LCD_1LINE;
    i2cdisplay_send_command(LCD_FUNCTIONSET | function);  
    
    // turn the display on with no cursor or blinking default
    i2cdisplay_set_display_control_mask(LCD_DISPLAYOFF | LCD_CURSOROFF | LCD_BLINKOFF);
    
    // clear display
    i2cdisplay_clear();

    // entry mode select
    i2cdisplay_set_display_mode_mask(LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT);

    // Turn display on
    i2cdisplay_display_on();
}

/********** high level commands, for the user! */


void i2cdisplay_display_on() {
    i2cdisplay_set_display_control_mask(display_control_mask | LCD_DISPLAYON);
}

void i2cdisplay_display_off() {
    i2cdisplay_set_display_control_mask(display_control_mask & ~LCD_DISPLAYON);
}

void i2cdisplay_curser_on() {
    i2cdisplay_set_display_control_mask(display_control_mask | LCD_CURSORON);
}

void i2cdisplay_curser_off() {
    i2cdisplay_set_display_control_mask(display_control_mask & ~LCD_CURSORON);
}

void i2cdisplay_blink_on() {
    i2cdisplay_set_display_control_mask(display_control_mask | LCD_BLINKON);
}

void i2cdisplay_blink_off() {
    i2cdisplay_set_display_control_mask(display_control_mask & ~LCD_BLINKON);
}

void i2cdisplay_clear(){
    i2cdisplay_send_command(LCD_CLEARDISPLAY);   // clear display, set cursor position to zero
    sleep(LCD_LONG_DELAY);                  // this command takes a long time!
}

void i2cdisplay_home(){
    i2cdisplay_send_command(LCD_RETURNHOME);     // set cursor position to zero
    sleep(LCD_LONG_DELAY);                  // this command takes a long time!
}

void i2cdisplay_set_cursor(uint8_t col, uint8_t row){
    uint8_t row_offsets[] = { 0x00, 0x40 }; //, 0x14, 0x54 };
    i2cdisplay_send_command(LCD_SETDDRAMADDR | (col + row_offsets[row]));
}

void i2cdisplay_backlight_off(void) {
    display_backlight_mask = LCD_NOBACKLIGHT;
    i2cdisplay_set_interface_pins(0x00);
}

void i2cdisplay_backlight_on(void) {
    display_backlight_mask = LCD_BACKLIGHT;
    i2cdisplay_set_interface_pins(0x00);
}

void i2cdisplay_write_text(const unsigned char *text) {
    uint8_t cur_row = 0;
    while (*text) {
        if (*text == '\n') {
            cur_row++;
            i2cdisplay_set_cursor(0, cur_row);
        } else {
            i2cdisplay_send_data(*text);            
        }
        *text++;
    }
}

void i2cdisplay_write_num(const uint16_t *num_ptr) {
    unsigned char buff[16]; 
    utoa(*num_ptr, buff, 10);
    i2cdisplay_write_text(buff);
}


/*********** mid level commands, for sending data/cmds */


void i2cdisplay_set_display_control_mask(uint8_t new_val) {
    display_control_mask = new_val;
    i2cdisplay_send_command(LCD_DISPLAYCONTROL | display_control_mask);  
}

void i2cdisplay_set_display_mode_mask(uint8_t new_val) {
    display_mode_mask = new_val;
    i2cdisplay_send_command(LCD_ENTRYMODESET | display_mode_mask);  
}

/************ low level data pushing commands **********/

inline void i2cdisplay_send_command(uint8_t value) {
    i2cdisplay_send(value, 0);
}

inline void i2cdisplay_send_data(uint8_t value) {
    i2cdisplay_send(value, Rs);
}

// write either command or data
void i2cdisplay_send(uint8_t value, uint8_t mode) {
    uint8_t highnib = value >> 4;
    uint8_t lownib = value & 0x0F;
    i2cdisplay_write4bits((highnib) | mode);
    i2cdisplay_write4bits((lownib) | mode);
    _delay_us(50);                              // commands need > 37us to settle
}

void i2cdisplay_write4bits(uint8_t value) {
    i2cdisplay_set_interface_pins(value | En);  // En high
    _delay_us(1);                               // enable pulse must be >450ns
    i2cdisplay_set_interface_pins(value & ~En); // En low
    _delay_us(1);                               // hold time
}

void i2cdisplay_set_interface_pins(uint8_t data) {  
    pcf8574_set_outputs(addr, (data | display_backlight_mask));
}

    